@@ -7,6 +7,7 @@
 # http://opensource.org/licenses/MIT>, at your option. This file may not be
 # copied, modified, or distributed except according to those terms.
 
+import os
 import sys
 
 import click
@@ -54,15 +55,73 @@ def raw(serial):
         sys.stdout.buffer.write(r)
 
 
+@click.command()
+@click.option("--count", default=64, help="How many bytes to generate (defaults to 8)")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
+def feedkernel(count, serial):
+    """Feed random bytes to /dev/random."""
+
+    if os.name != "posix":
+        print("This is a Linux-specific command!")
+        sys.exit(1)
+
+    if not 0 <= count <= 255:
+        print(f"Number of bytes must be between 0 and 255, you passed {count}")
+        sys.exit(1)
+
+    p = solo.client.find(serial)
+
+    import struct
+    import fcntl
+
+    RNDADDENTROPY = 0x40085203
+
+    entropy_info_file = "/proc/sys/kernel/random/entropy_avail"
+    print(f"Entropy before: 0x{open(entropy_info_file).read().strip()}")
+
+    r = p.get_rng(count)
+
+    # man 4 random
+
+    # RNDADDENTROPY
+    #       Add some additional entropy to the input pool, incrementing the
+    #       entropy count. This differs from writing to /dev/random or
+    #       /dev/urandom, which only adds some data but does not increment the
+    #       entropy count. The following structure is used:
+
+    #           struct rand_pool_info {
+    #               int    entropy_count;
+    #               int    buf_size;
+    #               __u32  buf[0];
+    #           };
+
+    #       Here entropy_count is the value added to (or subtracted from) the
+    #       entropy count, and buf is the buffer of size buf_size which gets
+    #       added to the entropy pool.
+
+    entropy_bits_per_byte = 2  # maximum 8, tend to be pessimistic
+    t = struct.pack(f"ii{count}s", count * entropy_bits_per_byte, count, r)
+
+    with open("/dev/random", mode="wb") as fh:
+        _ = fcntl.ioctl(fh, RNDADDENTROPY, t)
+    print(f"Entropy after:  0x{open(entropy_info_file).read().strip()}")
+
+
 @click.command()
 @click.option("-s", "--serial", help="Serial number of Solo use")
+@click.option(
+    "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
+)
 @click.argument("hash-type")
 @click.argument("filename")
-def probe(serial, hash_type, filename):
+def probe(serial, udp, hash_type, filename):
     """Calculate HASH."""
 
-    hash_type = hash_type.upper()
-    assert hash_type in ("SHA256", "SHA512")
+    if udp:
+        solo.fido2.force_udp_backend()
+
+    # hash_type = hash_type.upper()
+    assert hash_type in ("SHA256", "SHA512", "RSA2048", "Ed25519")
 
     data = open(filename, "rb").read()
     # < CTAPHID_BUFFER_SIZE
@@ -78,12 +137,32 @@ def probe(serial, hash_type, filename):
     from solo.commands import SoloBootloader
 
     result = p.send_data_hid(SoloBootloader.HIDCommandProbe, serialized_command)
-    print(result.hex())
+    result_hex = result.hex()
+    print(result_hex)
+    if hash_type == "Ed25519":
+        print(f"content: {result[64:]}")
+        # print(f"content from hex: {bytes.fromhex(result_hex[128:]).decode()}")
+        print(f"content from hex: {bytes.fromhex(result_hex[128:])}")
+        print(f"signature: {result[:128]}")
+        import nacl.signing
+
+        # verify_key = nacl.signing.VerifyKey(bytes.fromhex("c69995185efa20bf7a88139f5920335aa3d3e7f20464345a2c095c766dfa157a"))
+        verify_key = nacl.signing.VerifyKey(
+            bytes.fromhex(
+                "c69995185efa20bf7a88139f5920335aa3d3e7f20464345a2c095c766dfa157a"
+            )
+        )
+        try:
+            _ = verify_key.verify(result)
+            verified = True
+        except nacl.exceptions.BadSignatureError:
+            verified = False
+        print(f"verified? {verified}")
     # print(fido2.cbor.loads(result))
 
 
 # @click.command()
-# @click.option("-s", "--serial", help="Serial number of Solo use")
+# @click.option("-s", "--serial", help="Serial number of Solo to use")
 # @click.argument("filename")
 # def sha256sum(serial, filename):
 #     """Calculate SHA256 hash of FILENAME."""
@@ -97,7 +176,7 @@ def probe(serial, hash_type, filename):
 #     print(sha256sum.hex().lower())
 
 # @click.command()
-# @click.option("-s", "--serial", help="Serial number of Solo use")
+# @click.option("-s", "--serial", help="Serial number of Solo to use")
 # @click.argument("filename")
 # def sha512sum(serial, filename):
 #     """Calculate SHA512 hash of FILENAME."""
@@ -112,7 +191,7 @@ def probe(serial, hash_type, filename):
 
 
 @click.command()
-@click.option("-s", "--serial", help="Serial number of Solo use")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
 def reset(serial):
     """Reset key - wipes all credentials!!!"""
     if click.confirm(
@@ -124,7 +203,7 @@ def reset(serial):
 
 
 @click.command()
-@click.option("-s", "--serial", help="Serial number of Solo use")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
 @click.option(
     "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
 )
@@ -158,7 +237,7 @@ def verify(serial, udp):
 
 
 @click.command()
-@click.option("-s", "--serial", help="Serial number of Solo use")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
 @click.option(
     "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
 )
@@ -180,7 +259,7 @@ def version(serial, udp):
 
 
 @click.command()
-@click.option("-s", "--serial", help="Serial number of Solo use")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
 @click.option(
     "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
 )
@@ -195,6 +274,7 @@ def wink(serial, udp):
 key.add_command(rng)
 rng.add_command(hexbytes)
 rng.add_command(raw)
+rng.add_command(feedkernel)
 key.add_command(reset)
 key.add_command(update)
 key.add_command(probe)
